from typing import Optional, Dict, Any

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import ProgressLogMetric
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class FrameworkSqlTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        sql: Optional[str] = None,
        view: Optional[str] = None,
        log_sql: bool = False,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.logger = get_logger(__name__)

        if not sql:
            raise ValueError("sql is None or empty")

        if not view:
            raise ValueError("view is None or empty")

        self.sql: Param = Param(self, "sql", "")
        self._setDefault(sql=None)

        self.view: Param = Param(self, "view", "")
        self._setDefault(view=None)

        self.log_sql: Param = Param(self, "log_sql", "")
        self._setDefault(log_sql=False)

        self.verify_count_remains_same: Param = Param(
            self, "verify_count_remains_same", ""
        )
        self._setDefault(verify_count_remains_same=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyUnusedLocal,PyMissingOrEmptyDocstring,PyPep8Naming
    @keyword_only
    def setParams(
        self,
        sql: Optional[str] = None,
        view: Optional[str] = None,
        log_sql: bool = False,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        return self._set(**kwargs)  # type: ignore

    def _transform(self, df: DataFrame) -> DataFrame:
        sql_text: str = self.getSql()
        name: Optional[str] = self.getName()
        view: Optional[str] = self.getView()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        with ProgressLogMetric(
            name=name or view or "", progress_logger=progress_logger
        ):
            if progress_logger and name:
                # mlflow opens .txt files inline so we use that extension
                progress_logger.log_artifact(
                    key=f"{name}.sql.txt", contents=sql_text
                )
                progress_logger.write_to_log(name=name, message=sql_text)
            try:
                df = df.sql_ctx.sql(sql_text)
            except Exception:
                self.logger.info(f"Error in {name}")
                self.logger.info(sql_text)
                raise

            df.createOrReplaceTempView(view)
            self.logger.info(
                f"GenericSqlTransformer [{name}] finished running SQL"
            )

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setSql(self, value: str) -> 'FrameworkSqlTransformer':
        self._paramMap[self.sql] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSql(self) -> str:
        return self.getOrDefault(self.sql)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setView(self, value: str) -> 'FrameworkSqlTransformer':
        self._paramMap[self.view] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setLogSql(self, value: bool) -> 'FrameworkSqlTransformer':
        self._paramMap[self.log_sql] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLogSql(self) -> bool:
        return self.getOrDefault(self.log_sql)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setVerifyCountRemainsSame(
        self, value: bool
    ) -> 'FrameworkSqlTransformer':
        self._paramMap[self.verify_count_remains_same] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getVerifyCountRemainsSame(self) -> bool:
        return self.getOrDefault(  # type: ignore
            self.verify_count_remains_same
        )



from pathlib import Path
from typing import Any, Dict, List, Union, Optional

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param.shared import Param

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_local_file_loader.v1.framework_local_file_loader import \
    FrameworkLocalFileLoader


class FrameworkCsvLoader(FrameworkLocalFileLoader):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: str,
        path_to_csv: Union[str, List[str], Path],
        delimiter: str = ",",
        has_header: bool = True,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        **kwargs: Dict[Any, Any]
    ) -> None:
        super().__init__(
            view=view,
            filepath=path_to_csv,
            name=name,
            parameters=parameters,
            progress_logger=progress_logger,
            **kwargs
        )

        self.delimiter: Param = Param(self, "delimiter", "")
        self._setDefault(delimiter=",")

        self.has_header: Param = Param(self, "has_header", "")
        self._setDefault(has_header=True)

        self._set(delimiter=delimiter, has_header=has_header)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setDelimiter(self, value: Param) -> 'FrameworkCsvLoader':
        self._paramMap[self.delimiter] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDelimiter(self) -> str:
        return self.getOrDefault(self.delimiter)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setHasHeader(self, value: Param) -> 'FrameworkCsvLoader':
        self._paramMap[self.has_header] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHasHeader(self) -> bool:
        return self.getOrDefault(self.has_header)  # type: ignore

    def getReaderFormat(self) -> str:
        return "csv"

    def getReaderOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = {
            "header": "true" if self.getHasHeader() else "false",
            "delimiter": self.getDelimiter()
        }
        return options



import os
from json import loads, dumps
from pathlib import Path
from typing import Any, List


# noinspection SpellCheckingInspection
def convert_json_to_jsonl(src_file: Path, dst_file: Path) -> Path:
    """
    Converts json files to jsonl file
    :param src_file: json file
    :param dst_file:
    :return: path to jsonl file
    """
    assert os.path.isfile(src_file)

    # first detect if the src_file is already in jsonl format
    with open(src_file, "r") as file:
        beginning_lines: List[str] = file.readlines()
        # if each line begins with { then this is a jsonl file
        is_jsonl_file: bool = all(
            [line.lstrip().startswith("{") for line in beginning_lines]
        )

    if is_jsonl_file:  # already in jsonl format
        with open(src_file, "r") as file:
            # just copy the file over and be done
            contents: str = file.read()
            with open(dst_file, "w+") as file2:
                file2.write(contents)
            return dst_file

    # file is a json file so need to convert to jsonl
    with open(src_file, "r") as file:
        json_object: Any = loads(file.read())
    with open(dst_file, "w+") as file:
        if isinstance(json_object, list):
            for json_object_inner in json_object:
                json_text_inner: str = dumps(
                    obj=json_object_inner, separators=(',', ':')
                )
                file.write(json_text_inner)
                file.write("\n")
        else:
            json_text: str = dumps(obj=json_object, separators=(',', ':'))
            file.write(json_text)
            file.write("\n")

    return dst_file



import traceback
from typing import Any


class FriendlySparkException(Exception):
    # noinspection PyUnusedLocal
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        try:
            # Summary is a boolean argument
            # If True, it prints the exception summary
            # This way, we can avoid printing the summary all
            # the way along the exception "bubbling up"
            stage_name = kwargs['stage_name']
            error_text = (stage_name or ''
                          ) + ": " + FriendlySparkException.exception_summary()
            Exception.__init__(self, error_text)
        except KeyError:
            pass

    # noinspection SpellCheckingInspection
    @staticmethod
    def errortext(text: str) -> str:
        # Makes exception summary both BOLD and RED (FAIL)
        return text

    @staticmethod
    def exception_summary() -> str:
        # Gets the error stack
        msg = traceback.format_exc()
        try:
            # Builds the "frame" around the text
            # Gets the information about the error and makes it BOLD and RED
            info = list(
                filter(
                    lambda t: len(t) and t[0] != '\t',
                    msg.split('\n')[::-1]
                )
            )
            error = FriendlySparkException.errortext(
                'Error\t: {}'.format(info[0])
            )
            # Figure out where the error happened - location (file/notebook), line and function
            idx = [t.strip()[:4] for t in info].index('File')
            where = [v.strip() for v in info[idx].strip().split(',')]
            location, line, func = where[0][5:], where[1][5:], where[2][3:]
            # If it is a pyspark error, just go with it
            if 'pyspark' in error:
                new_msg = '\n{}'.format(error)
            # Otherwise, build the summary
            else:
                new_msg = '\nLocation: {}\nLine\t: {}\nFunction: {}\n{}'.format(
                    location, line, func, error
                )
            return new_msg
        except Exception as e:
            # If we managed to raise an exception while trying to format the original exception...
            # Oh, well...
            return 'This is awkward... \n{}'.format(str(e))



from typing import Dict, Any, List, Optional

from pyspark.ml.base import Transformer
from pyspark.ml.param import Param
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


class FrameworkTransformer(
    Transformer,  # type: ignore
    DefaultParamsReadable,  # type: ignore
    DefaultParamsWritable  # type: ignore
):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ):
        super(FrameworkTransformer, self).__init__()

        self.logger = get_logger(__name__)

        self.name: Param = Param(self, "name", "")
        self._setDefault(name=None)

        self.progress_logger: Param = Param(self, "progress_logger", "")
        self._setDefault(progress_logger=None)

        self.parameters: Param = Param(self, "parameters", "")
        self._setDefault(parameters=None)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    def setParams(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ) -> Any:
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        return df

    @property
    def transformers(self) -> List[Transformer]:
        return [self]

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setName(self, value: str) -> 'FrameworkTransformer':
        self._paramMap[self.name] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setProgressLogger(
        self, value: Optional[ProgressLogger]
    ) -> 'FrameworkTransformer':
        self._paramMap[self.progress_logger] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getProgressLogger(self) -> Optional[ProgressLogger]:
        return self.getOrDefault(self.progress_logger)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setParameters(
        self, value: Optional[Dict[str, Any]]
    ) -> 'FrameworkTransformer':
        self._paramMap[self.parameters] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getParameters(self) -> Optional[Dict[str, Any]]:
        return self.getOrDefault(self.parameters)  # type: ignore



from pathlib import Path
from typing import Callable, Dict, Any, Union, List

from pyspark.sql import SparkSession
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from spark_pipeline_framework.proxy_generator.python_transformer_helpers import get_python_function_from_location


def test_can_find_python_function(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath('./')

    # Act
    result_function: Callable[[Dict[str, Any]], Union[
        AutoMapperBase, List[AutoMapperBase]]] = (
            get_python_function_from_location(
                location=str(
                    data_dir.joinpath("library/features/carriers_mapping/v1")
                ),
                import_module_name='.mapping'
            )
        )

    result: Union[AutoMapperBase,
                  List[AutoMapperBase]] = result_function({"view": "bar"})

    # Assert
    assert isinstance(result, AutoMapperBase)



import os
from glob import glob
from pathlib import Path
from typing import Any, Dict, Optional, Union

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer
from spark_pipeline_framework.utilities.json_to_jsonl_converter import convert_json_to_jsonl


# noinspection SpellCheckingInspection
class FrameworkJsonToJsonlConverter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        file_path: Union[Path, str],
        output_folder: Union[Path, str],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ):
        """
        Convert json formats to jsonl since Sparks wants jsonl
        https://spark.apache.org/docs/latest/sql-data-sources-json.html
        https://jsonlines.org/


        :param file_path: path to original json file of folder containing json files
        :param output_folder: folder to put the jsonl files
        :param name:
        :param parameters:
        :param progress_logger:
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.file_path: Param = Param(self, "file_path", "")
        # noinspection Mypy
        self._setDefault(file_path=file_path)

        self.output_folder: Param = Param(self, "output_folder", "")
        # noinspection Mypy
        self._setDefault(output_folder=output_folder)

        # noinspection Mypy
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal,Mypy
    @keyword_only
    def setParams(
        self,
        file_path: Union[Path, str],
        output_folder: Union[Path, str],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ) -> Any:
        # noinspection Mypy
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        # noinspection Mypy
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        file_path: Union[Path, str] = self.getFilePath()
        output_folder: Union[Path, str] = self.getOutputFolder()
        # check if the file_path is a folder or a file
        if os.path.exists(file_path):
            if os.path.isdir(file_path):
                # read every file and convert it
                file_name: str
                for file_name in glob(str(Path(file_path).joinpath("*.json"))):
                    convert_json_to_jsonl(
                        src_file=Path(file_name),
                        dst_file=Path(output_folder).joinpath(
                            os.path.basename(file_name)
                        )
                    )
            elif os.path.isfile(file_path):
                # read this file and convert it
                convert_json_to_jsonl(
                    src_file=Path(file_path),
                    dst_file=Path(output_folder).joinpath(
                        os.path.basename(file_path)
                    )
                )
            else:
                raise NotImplementedError(
                    f"file_path: {file_path} is neither a folder nor a file"
                )
        else:
            raise FileNotFoundError(file_path)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[Path, str]:
        return self.getOrDefault(self.file_path)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOutputFolder(self) -> Union[Path, str]:
        return self.getOrDefault(self.output_folder)  # type: ignore



from typing import Dict, Any, List, Optional, Union

# noinspection PyProtectedMember
from pyspark import SparkContext, SQLContext, Row
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType, DataType


def create_view_from_dictionary(
    view: str,
    data: List[Dict[str, Any]],
    spark_session: SparkSession,
    schema: Optional[Union[DataType, str]] = None
) -> DataFrame:
    df: DataFrame = spark_session.createDataFrame(data=data, schema=schema)
    df.createOrReplaceTempView(name=view)
    return df


def create_empty_dataframe(spark_session: SparkSession) -> DataFrame:
    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    return df


def create_dataframe_from_json(
    spark_session: SparkSession, schema: StructType, json: str
) -> DataFrame:
    return spark_session.read.schema(schema).json(
        spark_session.sparkContext.parallelize([json])
    )


def spark_is_data_frame_empty(df: DataFrame) -> bool:
    """
    Efficient way to check if the data frame is empty without getting the count of the whole data frame
    """
    # from: https://stackoverflow.com/questions/32707620/how-to-check-if-spark-dataframe-is-empty
    return not bool(df.head(1))


def spark_get_execution_plan(df: DataFrame, extended: bool = False) -> Any:
    if extended:
        # noinspection PyProtectedMember
        return df._jdf.queryExecution().toString()
    else:
        # noinspection PyProtectedMember
        return df._jdf.queryExecution().simpleString()


def spark_table_exists(sql_ctx: SQLContext, view: str) -> bool:
    # noinspection PyBroadException
    return view in sql_ctx.tableNames()


def sc(df: DataFrame) -> SparkContext:
    # noinspection PyProtectedMember
    return df._sc


def add_metadata_to_column(
    df: DataFrame, column: str, metadata: Any
) -> DataFrame:
    return df.withColumn(column, df[column].alias(column, metadata=metadata))


def get_metadata_of_column(df: DataFrame, column: str) -> Any:
    return df.select(column).schema[0].metadata


def to_dicts(df: DataFrame, limit: int) -> List[Dict[str, Any]]:
    """
    converts a data frame into a list of dictionaries
    :param df:
    :param limit:
    :return:
    """
    row: Row
    rows = [row.asDict(recursive=True) for row in df.limit(limit).collect()]
    return rows



from typing import Optional, Dict, Any

from pyspark import keyword_only
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.python_proxy_base import PythonProxyBase


class FeatureCarrierPythonTransformer(PythonProxyBase):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        super().__init__(
            name=name,
            parameters=parameters,
            progress_logger=progress_logger,
            verify_count_remains_same=verify_count_remains_same
        )
        assert parameters
        assert progress_logger

    def _transform(self, df: DataFrame) -> DataFrame:
        pass



import pkgutil
import re
import sys
import types
from importlib import import_module
from inspect import signature
from typing import Dict, Any, Optional, Callable, Union, List, cast

from pyspark.ml import Transformer
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


def get_python_transformer_from_location(
    location: str, import_module_name: str, parameters: Dict[str, Any],
    progress_logger: Optional[ProgressLogger]
) -> Transformer:
    assert location
    assert isinstance(parameters, dict)
    assert progress_logger
    search = re.search(r'/library/', location)
    assert search
    lib_path = location[search.start() + 1:].replace('/', '.').replace('', '')
    # load_all_modules_from_dir(location)
    module = import_module(import_module_name, lib_path)
    md = module.__dict__
    my_class = [
        md[c] for c in md
        if (isinstance(md[c], type) and md[c].__module__ == module.__name__)
    ][0]
    my_class_signature = signature(my_class.__init__)
    my_class_args = [
        param.name for param in my_class_signature.parameters.values()
        if param.name != 'self'
    ]
    # now figure out the class_parameters to use when instantiating the class
    class_parameters = parameters.copy()
    class_parameters["parameters"] = parameters
    class_parameters['progress_logger'] = progress_logger
    if len(my_class_args) > 0 and len(class_parameters) > 0:
        return cast(
            Transformer,
            my_class(
                **{
                    k: v
                    for k, v in class_parameters.items() if k in my_class_args
                }
            )
        )
    else:
        return cast(Transformer, my_class())


def get_python_function_from_location(
    location: str, import_module_name: str
) -> Callable[[Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]]:
    assert location
    search = re.search(r'/library/', location)
    assert search
    lib_path = location[search.start() + 1:].replace('/', '.').replace('', '')
    # load_all_modules_from_dir(location)
    module = import_module(import_module_name, lib_path)
    md = module.__dict__
    # noinspection PyTypeChecker
    my_function: Callable[[Dict[str, Any]],
                          Union[AutoMapperBase, List[AutoMapperBase]]] = [
                              md[c] for c in md
                              if isinstance(md[c], types.FunctionType)
                          ][0]
    my_function_signature = signature(my_function)
    my_function_args = [
        param for param in my_function_signature.parameters.values()
        if param.name != 'self'
    ]
    # now figure out the function_parameters has one and only one parameter: parameters
    assert len(my_function_args) == 1
    assert my_function_args[0].name == "parameters"
    return my_function


def load_all_modules_from_dir(dirname: str) -> None:
    for importer, package_name, _ in pkgutil.iter_modules([dirname]):
        full_package_name = '%s.%s' % (dirname, package_name)
        if full_package_name not in sys.modules:
            module = importer.find_module(package_name
                                          ).load_module(full_package_name)
            print(module)



from typing import List, Any


def flatten(my_list: List[Any]) -> List[Any]:
    """

    :param my_list:
    :return:
    """
    if not my_list:
        return my_list
    if isinstance(my_list[0], list):
        return flatten(my_list[0]) + flatten(my_list[1:])
    return my_list[:1] + flatten(my_list[1:])


# def create_stages(transformer_list: List, parameters, progress_logger):
#     return [t(parameters, progress_logger) for t in flatten(transformer_list)]



from pyspark.ml import Transformer
from pyspark.sql import SparkSession

from library.features.carriers_python.v1.calculate import FeatureCarrierPythonTransformer
from library.features.carriers_python.v1.features_carriers_python_v1 import FeaturesCarriersPythonV1
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.spark_data_frame_helpers import create_empty_dataframe


def test_can_run_python_transformer(spark_session: SparkSession) -> None:
    # Arrange

    # Act
    with ProgressLogger() as progress_logger:
        result: Transformer = FeaturesCarriersPythonV1(
            parameters={
                "foo": "bar"
            }, progress_logger=progress_logger
        ).transformers[0]

    # Assert
    assert isinstance(result, FeatureCarrierPythonTransformer)

    # make sure we can call transform on it
    df = create_empty_dataframe(spark_session=spark_session)
    result.transform(df)



import logging
import os
import shutil

# noinspection PyPackageRequirements
from typing import Any

import pytest
from pyspark.sql import SparkSession

# make sure env variables are set correctly
if 'SPARK_HOME' not in os.environ:
    os.environ['SPARK_HOME'] = '/usr/local/opt/spark'


def quiet_py4j() -> None:
    """ turn down spark logging for the carriers context """
    logger = logging.getLogger('py4j')
    logger.setLevel(logging.ERROR)


def clean_spark_dir() -> None:
    """

    :return:
    """
    try:
        os.remove("./derby.log")
        shutil.rmtree("./metastore_db")
        shutil.rmtree("./spark-warehouse")
    except OSError:
        pass


def clean_spark_session(session: SparkSession) -> None:
    """

    :param session:
    :return:
    """
    tables = session.catalog.listTables("default")

    for table in tables:
        print(f"clear_tables() is dropping table/view: {table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP TABLE IF EXISTS default.{table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP VIEW IF EXISTS default.{table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP VIEW IF EXISTS {table.name}")

    session.catalog.clearCache()


def clean_close(session: SparkSession) -> None:
    """

    :param session:
    :return:
    """
    clean_spark_session(session)
    clean_spark_dir()
    session.stop()


@pytest.fixture(scope="session")
def spark_session(request: Any) -> SparkSession:
    # make sure env variables are set correctly
    if 'SPARK_HOME' not in os.environ:
        os.environ['SPARK_HOME'] = '/usr/local/opt/spark'

    clean_spark_dir()

    master = "local[2]"

    session = SparkSession.builder.appName("pytest-pyspark-local-testing"). \
        master(master). \
        config("spark.ui.showConsoleProgress", "false"). \
        config("spark.sql.shuffle.partitions", "2"). \
        config("spark.default.parallelism", "4"). \
        config("spark.sql.broadcastTimeout", "2400"). \
        enableHiveSupport(). \
        getOrCreate()

    request.addfinalizer(lambda: clean_close(session))
    quiet_py4j()
    return session



from pathlib import Path

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType
from tests.conftest import clean_spark_session

from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import FrameworkCsvLoader


def assert_results(result: DataFrame) -> None:
    """
    Shared asserts for the different formats of CSV file, all of which contain the same data.
    """
    # Assert
    assert result.count() == 3

    assert result.collect()[1][0] == "2"
    assert result.collect()[1][1] == "bar"
    assert result.collect()[1][2] == "bar2"


# noinspection SqlNoDataSourceInspection
def test_can_load_simple_csv(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('test.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkCsvLoader(
        view="my_view", path_to_csv=test_file_path, delimiter=","
    ).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM my_view")

    result.show()

    # Assert
    assert_results(result)


# noinspection SqlNoDataSourceInspection
def test_can_load_non_standard_delimited_csv(
    spark_session: SparkSession
) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('test.psv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    loader = FrameworkCsvLoader(
        view="my_view", path_to_csv=test_file_path, delimiter="|"
    )
    loader.transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM my_view")

    result.show()

    # Assert
    assert loader.getDelimiter() == "|"
    assert_results(result)


# noinspection SqlNoDataSourceInspection
def test_can_load_csv_without_header(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('no_header.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkCsvLoader(
        view="another_view",
        path_to_csv=test_file_path,
        delimiter=",",
        has_header=False
    ).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM another_view")

    # Assert
    assert_results(result)



from pathlib import Path
from typing import Callable, Dict, Any, Union, List

from pyspark.sql import SparkSession, DataFrame
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.proxy_generator.python_transformer_helpers import get_python_function_from_location
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import FrameworkMappingLoader


def test_can_run_framework_mapping_runner(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.parent.joinpath('./')

    spark_session.createDataFrame(
        [
            (1, 'Qureshi', 'Imran'),
            (2, 'Vidal', 'Michael'),
        ], ['member_id', 'last_name', 'first_name']
    ).createOrReplaceTempView("patients")

    source_df: DataFrame = spark_session.table("patients")

    df = source_df.select("member_id")
    df.createOrReplaceTempView("members")

    # Act
    mapping_function: Callable[[Dict[str, Any]], Union[
        AutoMapperBase, List[AutoMapperBase]]] = (
            get_python_function_from_location(
                location=str(
                    data_dir.joinpath("library/features/carriers_mapping/v1")
                ),
                import_module_name='.mapping'
            )
        )
    with ProgressLogger() as progress_logger:
        FrameworkMappingLoader(
            view="members",
            mapping_function=mapping_function,
            parameters={
                "foo": "bar"
            },
            progress_logger=progress_logger
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("members")

    result_df.printSchema()
    result_df.show(truncate=False)

    assert len(result_df.columns) == 5
    assert result_df.where("patient_id == 1"
                           ).select("dst1").collect()[0][0] == "src1"
    assert result_df.where("patient_id == 1"
                           ).select("dst2").collect()[0][0][0] == "address1"

    assert result_df.where("patient_id == 1"
                           ).select("dst3").collect()[0][0][0] == "address1"
    assert result_df.where("patient_id == 1"
                           ).select("dst3").collect()[0][0][1] == "address2"

    assert result_df.where("patient_id == 1"
                           ).select("dst4").collect()[0][0][0][0] == "usual"
    assert result_df.where("patient_id == 1"
                           ).select("dst4").collect()[0][0][0][1] == "Qureshi"



from typing import Dict, Any, Callable, Optional, Union, List

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql import DataFrame
from spark_auto_mapper.automappers.automapper import AutoMapper
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class FrameworkMappingLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: str,
        mapping_function: Callable[[Dict[str, Any]],
                                   Union[AutoMapperBase,
                                         List[AutoMapperBase]]],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ) -> None:
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param = Param(self, "view", "")
        # noinspection Mypy
        self._setDefault(view=None)

        self.mapping_function: Callable[[Dict[str, Any]], Union[
            AutoMapperBase, List[AutoMapperBase]]] = mapping_function

        # noinspection Mypy
        kwargs = self._input_kwargs
        # remove mapping_function since that is not serializable
        kwargs = {
            key: value
            for key, value in kwargs.items() if key != "mapping_function"
        }
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal,Mypy
    @keyword_only
    def setParams(
        self,
        view: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ) -> Any:
        # noinspection Mypy
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        # noinspection Mypy
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        mapping_function: Callable[[Dict[str, Any]], Union[
            AutoMapperBase, List[AutoMapperBase]]] = self.getMappingFunction()

        # run the mapping function to get an AutoMapper
        incoming_parameters: Optional[Dict[str, Any]] = self.getParameters()
        parameters: Dict[
            str,
            Any] = incoming_parameters.copy() if incoming_parameters else {}
        parameters["view"] = view

        auto_mapper: Union[
            AutoMapperBase,
            List[AutoMapperBase]] = mapping_function(parameters)

        assert isinstance(auto_mapper, AutoMapper) or (auto_mapper, list)

        if isinstance(auto_mapper, AutoMapper):
            # then call transform() on the AutoMapper
            df = auto_mapper.transform(df=df)
            # create the view
            df.createOrReplaceTempView(view)
        elif isinstance(auto_mapper, list):
            auto_mappers: List[AutoMapperBase] = auto_mapper
            for a in auto_mappers:
                assert isinstance(a, AutoMapper)
                a.transform(df=df)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMappingFunction(
        self
    ) -> Callable[[Dict[str, Any]], Union[AutoMapperBase,
                                          List[AutoMapperBase]]]:
        return self.mapping_function



from pathlib import Path
from typing import Union, List

from pyspark.ml.base import Transformer, Estimator
from pyspark.ml.feature import SQLTransformer
from pyspark.ml.pipeline import Pipeline
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import FrameworkCsvLoader


def test_simple_csv_loader_pipeline(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath('./')
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # noinspection SqlDialectInspection,SqlNoDataSourceInspection
    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    # parameters = Dict[str, Any]({
    # })

    stages: List[Union[Estimator, Transformer]] = [
        FrameworkCsvLoader(view="flights", path_to_csv=flights_path),
        SQLTransformer(statement="SELECT * FROM flights"),
    ]

    pipeline: Pipeline = Pipeline(stages=stages)

    transformer = pipeline.fit(df)
    result_df: DataFrame = transformer.transform(df)

    # Assert
    result_df.show()

    assert result_df.count() > 0



import shutil
from os import path
from pathlib import Path

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import FrameworkCsvLoader
from spark_pipeline_framework.transformers.framework_parquet_exporter.v1.framework_parquet_exporter import \
    FrameworkParquetExporter
from spark_pipeline_framework.transformers.framework_parquet_loader.v1.framework_parquet_loader import \
    FrameworkParquetLoader
from tests.spark_test_helper import SparkTestHelper


def test_can_save_parquet(spark_session: SparkSession) -> None:
    # Arrange
    SparkTestHelper.clear_tables(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('test.csv')}"

    if path.isdir(data_dir.joinpath('temp')):
        shutil.rmtree(data_dir.joinpath('temp'))

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    FrameworkCsvLoader(
        view="my_view", path_to_csv=test_file_path, delimiter=","
    ).transform(df)

    parquet_file_path: str = f"file://{data_dir.joinpath('temp/').joinpath(f'test.parquet')}"

    # Act
    FrameworkParquetExporter(view="my_view",
                             file_path=parquet_file_path).transform(df)

    # Assert
    FrameworkParquetLoader(view="my_view2",
                           file_path=parquet_file_path).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM my_view2")

    result.show()

    assert result.count() == 3

    assert result.collect()[1][0] == "2"
    assert result.collect()[1][1] == "bar"
    assert result.collect()[1][2] == "bar2"



from pathlib import Path
from typing import List, Union, Dict, Any, Optional

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.utils import AnalysisException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import ProgressLogMetric
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class FrameworkParquetLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: str,
        file_path: Union[str, List[str], Path],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        merge_schema: bool = False,
        limit: int = -1
    ):
        super(FrameworkParquetLoader, self).__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.merge_schema: Param = Param(self, "merge_schema", "")
        self._setDefault(merge_schema=None)

        self.limit: Param = Param(self, "limit", "")
        self._setDefault(limit=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    @keyword_only
    def setParams(
        self,
        view: str,
        file_path: Union[str, List[str], Path],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        merge_schema: bool = False,
        limit: int = -1
    ) -> None:
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        return self._set(**kwargs)  # type: ignore

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        path: Union[str, List[str], Path] = self.getFilePath()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        merge_schema: bool = self.getMergeSchema()
        limit: int = self.getLimit()

        with ProgressLogMetric(
            name=f"{name or view}_table_loader",
            progress_logger=progress_logger
        ):
            try:
                if merge_schema is True:
                    final_df = df.sql_ctx.read.option(
                        "mergeSchema", "true"
                    ).format("parquet").load(path=str(path))
                else:
                    final_df = df.sql_ctx.read.format("parquet").load(
                        path=str(path)
                    )

                if limit and limit > 0:
                    final_df = final_df.limit(limit)

                # store new data frame in the view
                final_df.createOrReplaceTempView(view)
            except AnalysisException as e:
                self.logger.error(
                    f"File load failed. Location: {path} may be empty"
                )
                raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setView(self, value: str) -> 'FrameworkParquetLoader':
        self._paramMap[self.view] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setFilePath(
        self, value: Union[str, List[str], Path]
    ) -> 'FrameworkParquetLoader':
        self._paramMap[self.file_path] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[str, List[str], Path]:
        return self.getOrDefault(self.file_path)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setMergeSchema(self, value: bool) -> 'FrameworkParquetLoader':
        self._paramMap[self.merge_schema] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMergeSchema(self) -> bool:
        return self.getOrDefault(self.merge_schema)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setLimit(self, value: int) -> 'FrameworkParquetLoader':
        self._paramMap[self.limit] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)  # type: ignore



from logging import Logger
from pathlib import Path
from typing import Union, List, Optional, Dict, Any

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param.shared import Param
from pyspark.sql import DataFrame
from pyspark.sql.functions import input_file_name
from pyspark.sql.readwriter import DataFrameReader
from pyspark.sql.types import StructType

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class FrameworkLocalFileLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    # keyword_only: A decorator that forces keyword arguments in the wrapped method
    #     and saves actual input keyword arguments in `_input_kwargs`.
    @keyword_only
    def __init__(
        self,
        view: str,
        filepath: Union[str, List[str], Path],
        delimiter: str = ",",
        limit: int = -1,
        has_header: bool = True,
        infer_schema: bool = False,
        cache_table: bool = True,
        schema: StructType = None,
        create_file_path: bool = False,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        **kwargs: Dict[Any, Any]
    ) -> None:
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.logger: Logger = get_logger(__name__)

        self.view: Param = Param(self, "view", "")
        self._setDefault(view=None)

        self.filepath: Param = Param(self, "filepath", "")
        self._setDefault(filepath=None)

        self.schema: Param = Param(self, "schema", "")
        self._setDefault(schema=None)

        self.cache_table: Param = Param(self, "cache_table", "")
        self._setDefault(cache_table=True)

        self.limit: Param = Param(self, "limit", "")
        self._setDefault(limit=-1)

        self.infer_schema: Param = Param(self, "infer_schema", "")
        self._setDefault(infer_schema=False)

        self.create_file_path: Param = Param(self, "create_file_path", "")
        self._setDefault(create_file_path=False)

        if not filepath:
            raise ValueError("filepath is None or empty")

        self.logger.info(f"Received filepath: {filepath}")

        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view = self.getView()
        filepath: Union[str, List[str], Path] = self.getFilepath()
        schema = self.getSchema()
        cache_table = self.getCacheTable()
        infer_schema = self.getInferSchema()
        limit = self.getLimit()
        create_file_path = self.getCreateFilePath()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        if not filepath:
            raise ValueError(f"filepath is empty: {filepath}")

        if isinstance(filepath, str):
            if filepath.__contains__(":"):
                absolute_paths: List[str] = [filepath]
            else:
                data_dir = Path(__file__).parent.parent.joinpath('./')
                if isinstance(filepath, list):
                    absolute_paths = [
                        f"file://{data_dir.joinpath(path)}"
                        for path in filepath
                    ]
                else:
                    absolute_paths = [f"file://{data_dir.joinpath(filepath)}"]
        elif isinstance(filepath, Path):
            data_dir = Path(__file__).parent.parent.joinpath('./')
            absolute_paths = [f"file://{data_dir.joinpath(filepath)}"]
        else:
            data_dir = Path(__file__).parent.parent.joinpath('./')
            absolute_paths = [
                f"file://{data_dir.joinpath(path)}" for path in filepath
            ]

        progress_logger and progress_logger.write_to_log(
            f"Loading {self.getReaderFormat()} file for view {view}: {absolute_paths}, infer_schema: {infer_schema}"
        )

        df_reader: DataFrameReader = df.sql_ctx.read

        if schema:
            df_reader = df_reader.schema(schema)
        elif infer_schema:
            df_reader = df_reader.option("inferSchema", "true")
        else:
            df_reader = df_reader.option("inferSchema", "false")

        df2: DataFrame
        df_reader = df_reader.format(self.getReaderFormat())
        for k, v in self.getReaderOptions().items():
            df_reader = df_reader.option(k, v)

        if create_file_path:
            df2 = df_reader.load(absolute_paths
                                 ).withColumn("file_path", input_file_name())
        else:
            df2 = df_reader.load(absolute_paths)

        if limit and limit > -1:
            df2 = df2.limit(limit)

        # now save the data frame into view
        df2.createOrReplaceTempView(view)

        if cache_table:
            self.logger.info(f"Caching table {view}")
            df.sql_ctx.sql(f"CACHE TABLE {view}")

        progress_logger and progress_logger.write_to_log(
            f"Finished Loading {self.getReaderFormat()} file for View[{view}]: {absolute_paths}, "
            + f"infer_schema: {infer_schema}"
        )

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setView(self, value: Param) -> 'FrameworkLocalFileLoader':
        self._paramMap[self.view] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setFilepath(self, value: Param) -> 'FrameworkLocalFileLoader':
        self._paramMap[self.filepath] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilepath(self) -> Union[str, List[str], Path]:
        return self.getOrDefault(self.filepath)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setSchema(self, value: Param) -> 'FrameworkLocalFileLoader':
        self._paramMap[self.schema] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSchema(self) -> StructType:
        return self.getOrDefault(self.schema)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setCacheTable(self, value: Param) -> 'FrameworkLocalFileLoader':
        self._paramMap[self.cache_table] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getCacheTable(self) -> bool:
        return self.getOrDefault(self.cache_table)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setInferSchema(self, value: Param) -> 'FrameworkLocalFileLoader':
        self._paramMap[self.infer_schema] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getInferSchema(self) -> bool:
        return self.getOrDefault(self.infer_schema)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setLimit(self, value: Param) -> 'FrameworkLocalFileLoader':
        self._paramMap[self.limit] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setCreateFilePath(self, value: bool) -> 'FrameworkLocalFileLoader':
        self._paramMap[self.create_file_path] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getCreateFilePath(self) -> bool:
        return self.getOrDefault(self.create_file_path)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> str:
        return self.getView()

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderFormat(self) -> str:
        raise NotImplementedError("Must implement in baseclass")

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderOptions(self) -> Dict[str, Any]:
        raise NotImplementedError("Must implement in baseclass")



from typing import Optional, Dict, Any

from pyspark import keyword_only
from pyspark.ml.base import Transformer
from pyspark.ml.param import Param
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


class PythonProxyBase(
    Transformer,  # type: ignore
    DefaultParamsReadable,  # type: ignore
    DefaultParamsWritable  # type: ignore
):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        super(PythonProxyBase, self).__init__()

        self.name: Param = Param(self, "name", "")
        self._setDefault(name=None)

        self.parameters: Param = Param(self, "parameters", "")
        self._setDefault(parameters=None)

        self.progress_logger: Param = Param(self, "progress_logger", "")
        self._setDefault(progress_logger=None)

        self.verify_count_remains_same: Param = Param(
            self, "verify_count_remains_same", ""
        )
        self._setDefault(verify_count_remains_same=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyUnusedLocal
    @keyword_only
    def setParams(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> Any:
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setName(self, value: str) -> 'PythonProxyBase':
        self._paramMap[self.name] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> str:
        return self.getOrDefault(  # type: ignore
            self.name or self.__class__.__name__
        )

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setProgressLogger(self, value: ProgressLogger) -> 'PythonProxyBase':
        self._paramMap[self.progress_logger] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getProgressLogger(self) -> Optional[ProgressLogger]:
        return self.getOrDefault(self.progress_logger)  # type: ignore



from pyspark.sql.session import SparkSession

from spark_pipeline_framework.transformers.framework_jdbc_exporter.v1.framework_jdbc_exporter import \
    FrameworkJdbcExporter


def test_can_save_via_jdbc(spark_session: SparkSession) -> None:
    """
    Because testing against a database is a pain and most of the core transformation logic is in the base exporter,
    we're only testing that the options and format are correctly being exposed.
    """
    # Arrange
    view = "my_view"
    jdbc_url = "jdbc:mysql:user@password:host/db:port"
    table = "my_view_table"
    driver = "org.driver.FakeDriver"

    # Act
    exporter = FrameworkJdbcExporter(
        view=view,
        jdbc_url=jdbc_url,
        table=table,
        driver=driver,
        mode=FrameworkJdbcExporter.MODE_OVERWRITE,
    )

    # Assert
    options = exporter.getOptions()
    assert options["url"] == jdbc_url
    assert options["driver"] == driver
    assert exporter.getFormat() == "jdbc"
    assert exporter.getMode() == "overwrite"



from pyspark.sql.session import SparkSession


class SparkTestHelper:
    @staticmethod
    def clear_tables(spark_session: SparkSession) -> None:
        """
        :param spark_session:
        """

        # spark_session.sql("SET -v").show(n=200, truncate=False)

        tables = spark_session.catalog.listTables("default")

        for table in tables:
            print(f"clear_tables() dropping table/view: {table.name}")
            spark_session.sql(f"DROP TABLE IF EXISTS default.{table.name}")
            spark_session.sql(f"DROP VIEW IF EXISTS default.{table.name}")
            spark_session.sql(f"DROP VIEW IF EXISTS {table.name}")

        spark_session.catalog.clearCache()



from datetime import datetime
from types import TracebackType
from typing import Optional

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


class ProgressLogMetric:
    def __init__(self, name: str, progress_logger: Optional[ProgressLogger]):
        self.progress_logger: Optional[ProgressLogger] = progress_logger
        self.name: str = name
        self.start_time: datetime = datetime.now()

    def __enter__(self) -> 'ProgressLogMetric':
        return self.start()

    def __exit__(
        self, exc_type: Optional[BaseException],
        exc_value: Optional[BaseException], traceback: Optional[TracebackType]
    ) -> None:
        self.stop()

    def start(self) -> 'ProgressLogMetric':
        """
        start
        :return:
        """
        self.start_time = datetime.now()
        return self

    def stop(self) -> None:
        """
        stop
        """
        if self.name and self.progress_logger:
            end_time: datetime = datetime.now()
            time_diff_in_minutes: float = (end_time - self.start_time
                                           ).total_seconds() // 60
            if self.progress_logger:
                self.progress_logger.log_metric(
                    self.name, time_diff_in_minutes
                )



from os import getcwd
from os import walk, path, listdir
from re import search
from shutil import rmtree
from sys import exit
from typing import Match, Optional, List

# data_dir: Path = Path(__file__).parent.joinpath('./')
#
# PACKAGE_PARENT = '..'
# SCRIPT_DIR = path.dirname(path.realpath(data_dir))
# print(SCRIPT_DIR)
# normpath = path.normpath(path.join(SCRIPT_DIR, PACKAGE_PARENT))
# print(normpath)
# sys.path.append(normpath)


# TODO: This is copied here from proxy_generator.py.  Figure out how to import it
class ProxyGenerator:
    @staticmethod
    def remove_empty_dirs(folder: str) -> None:
        def is_path_empty(directory: str) -> bool:
            for dir_tuple in walk(directory):
                if [
                    file for file in dir_tuple[2] if not (
                        file.startswith(('__', '.')) or file.endswith('.pyc')
                    )
                ]:
                    return False
            return True

        for path_tuple in walk(folder):
            # noinspection SpellCheckingInspection
            if path_tuple[0].endswith('__pycache__'):
                pass
            else:
                current_path: str = path.join(folder, path_tuple[0])
                if is_path_empty(current_path):
                    try:
                        rmtree(current_path)
                    except FileNotFoundError:
                        pass
                    else:
                        print(f'Removed {current_path}')

    @staticmethod
    def generate_proxies(folder: str) -> None:
        all_objects_in_path: List[str] = listdir(folder)
        non_special_objects: List[str] = [
            file for file in all_objects_in_path
            if not file.startswith(('_', '.'))
        ]
        # noinspection SpellCheckingInspection
        folders: List[str] = [
            subfolder for subfolder in non_special_objects
            if '.' not in subfolder
            and len(listdir(path.join(folder, subfolder))) != 0
        ]
        files: List[str] = [
            file for file in all_objects_in_path
            if '.' in file and not file.startswith('_')
        ]
        transformer_file_indicators = ('.sql', '.csv', '.py')
        path_contains_transformer: bool = len(
            [
                file for file in files
                if file.endswith(transformer_file_indicators)
            ]
        ) > 0

        if path_contains_transformer:
            for file in files:
                if file.endswith(transformer_file_indicators):
                    search_result: Optional[Match[str]
                                            ] = search(r'/library/', folder)
                    if search_result:
                        transformer_reader_file_name = folder[
                            search_result.end():].replace('/', '_')
                        ProxyGenerator.write_transformer(
                            file_name=transformer_reader_file_name,
                            folder=folder
                        )

        # now recursively generate proxies
        # noinspection SpellCheckingInspection
        for subfolder in folders:
            ProxyGenerator.generate_proxies(path.join(folder, subfolder))

    @staticmethod
    def write_transformer(file_name: str, folder: str) -> None:
        transformer_reader_class_name = ''.join(
            [s.title() for s in file_name.split('_')]
        )
        transformer_reader_string = f"""
from typing import Optional, Dict, Any
from spark_pipeline_framework.proxy_generator.proxy_base import ProxyBase
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from os import path


# This file was auto-generated by generate_proxies(). It enables auto-complete in PyCharm. Do not edit manually!
class {transformer_reader_class_name}(ProxyBase):
    def __init__(self,
                 parameters: Dict[str, Any],
                 progress_logger: Optional[ProgressLogger] = None,
                 verify_count_remains_same: bool = False
                 ) -> None:
        location: str = path.dirname(path.abspath(__file__))
        super().__init__(
            parameters=parameters,
            location=location,
            progress_logger=progress_logger,
            verify_count_remains_same=verify_count_remains_same
        )
"""
        transformer_proxy_file_name: str = path.join(folder, file_name + '.py')
        if not path.exists(transformer_proxy_file_name):
            with open(transformer_proxy_file_name, 'w+') as file:
                print(f"Creating {transformer_proxy_file_name}")
                file.write(transformer_reader_string)


def main() -> int:
    library_folder: str = path.join(getcwd(), 'library')
    ProxyGenerator.remove_empty_dirs(library_folder)
    ProxyGenerator.generate_proxies(library_folder)
    return 0


if __name__ == "__main__":
    exit(main())



from os import path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import List

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_json_to_jsonl_converter.v1.framework_json_to_jsonl_converter import \
    FrameworkJsonToJsonlConverter
from tests.conftest import clean_spark_session


# noinspection SqlNoDataSourceInspection
def test_can_convert_json_folder_to_jsonl(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('test_files')}"

    temp_folder = data_dir.joinpath('temp')
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkJsonToJsonlConverter(
        file_path=test_file_path, output_folder=temp_folder
    ).transform(df)

    # Assert
    with open(temp_folder.joinpath("test.json"), "r+") as file:
        lines: List[str] = file.readlines()
        assert len(lines) == 2
        assert lines[
            0
        ] == '{"title":"A Philosophy of Software Design","authors":[{"given":["John"],"surname":"Ousterhout"}],"edition":null}\n'
        assert lines[
            1
        ] == '{"title":"Essentials of Programming Languages","authors":[{"given":["Dan","P."],"surname":"Friedman"},{"given":["Mitchell"],"surname":"Wand"}],"edition":3}\n'



from typing import Optional, Dict, Any
from spark_pipeline_framework.proxy_generator.proxy_base import ProxyBase
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from os import path


# This file was auto-generated by generate_proxies(). It enables auto-complete in PyCharm. Do not edit manually!
class FeaturesCarriersMultipleMappingsV1(ProxyBase):
    def __init__(
        self,
        parameters: Dict[str, Any],
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        location: str = path.dirname(path.abspath(__file__))
        super().__init__(
            parameters=parameters,
            location=location,
            progress_logger=progress_logger,
            verify_count_remains_same=verify_count_remains_same
        )



from typing import Any, Dict, Optional

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql import DataFrameWriter
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.utils import AnalysisException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import ProgressLogMetric
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class FrameworkBaseExporter(FrameworkTransformer):
    MODE_APPEND = "append"
    MODE_OVERWRITE = "overwrite"
    MODE_IGNORE = "ignore"
    MODE_ERROR = "error"
    MODE_CHOICES = (
        MODE_APPEND,
        MODE_OVERWRITE,
        MODE_IGNORE,
        MODE_ERROR,
    )

    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: Optional[str] = None,
        name: Optional[str] = None,
        mode: str = MODE_ERROR,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: int = -1
    ):
        assert mode in FrameworkBaseExporter.MODE_CHOICES

        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param = Param(self, "view", "")
        self._setDefault(view=view)

        self.mode: Param = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.limit: Param = Param(self, "limit", "")
        self._setDefault(limit=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    @keyword_only
    def setParams(
        self,
        view: Optional[str] = None,
        name: Optional[str] = None,
        mode: str = MODE_ERROR,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: int = -1
    ) -> None:
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        return self._set(**kwargs)  # type: ignore

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        name: Optional[str] = self.getName()
        format: str = self.getFormat()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        # limit: int = self.getLimit()

        with ProgressLogMetric(
            name=f"{name or view}_{format}_exporter",
            progress_logger=progress_logger
        ):
            try:
                writer: DataFrameWriter
                if view:
                    writer = df.sql_ctx.table(view).write.format(format)
                else:
                    writer = df.write.format(format)

                writer = writer.mode(self.getMode())

                for k, v in self.getOptions().items():
                    writer.option(k, v)

                writer.save()

            except AnalysisException as e:
                self.logger.error(f"Failed to write to {format}")
                raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setView(self, value: Param) -> 'FrameworkBaseExporter':
        self._paramMap[self.view] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setMode(self, value: Param) -> 'FrameworkBaseExporter':
        self._paramMap[self.mode] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setLimit(self, value: Param) -> 'FrameworkBaseExporter':
        self._paramMap[self.limit] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)  # type: ignore

    def getFormat(self) -> str:
        raise NotImplementedError

    def getOptions(self) -> Dict[str, Any]:
        raise NotImplementedError



from pathlib import Path

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType
from tests.conftest import clean_spark_session

from spark_pipeline_framework.transformers.framework_json_loader.v1.framework_json_loader import FrameworkJsonLoader


# noinspection SqlNoDataSourceInspection
def test_can_load_simple_json(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('test.json')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkJsonLoader(view="books", filepath=test_file_path,
                        multiLine=True).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM books")

    result.show()

    # Assert
    assert result.count() == 2

    assert result.collect(
    )[1]["title"] == "Essentials of Programming Languages"
    assert len(result.collect()[1]["authors"]) == 2
    assert result.collect()[1]["authors"][0]["surname"] == "Friedman"
    assert result.collect()[1]["edition"] == 3


# noinspection SqlNoDataSourceInspection
def test_can_load_newline_delimited_json(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('test.jsonl')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    loader = FrameworkJsonLoader(
        view="books", filepath=test_file_path, limit=100
    )
    loader.transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM books")

    result.show()

    # Assert
    assert result.count() == 2
    assert loader.getLimit() == 100

    assert result.collect(
    )[1]["title"] == "Essentials of Programming Languages"
    assert len(result.collect()[1]["authors"]) == 2
    assert result.collect()[1]["authors"][0]["surname"] == "Friedman"
    assert result.collect()[1]["edition"] == 3



from pathlib import Path

from pyspark.ml import Transformer
from pyspark.sql import SparkSession

from library.features.carriers_python.v1.calculate import FeatureCarrierPythonTransformer
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.python_transformer_helpers import get_python_transformer_from_location
from spark_pipeline_framework.utilities.spark_data_frame_helpers import create_empty_dataframe


def test_can_find_python_transformer(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath('./')

    # load_all_modules_from_dir(str(data_dir.joinpath("../../library/features/carriers_python/v1")))

    # Act
    with ProgressLogger() as progress_logger:
        result: Transformer = get_python_transformer_from_location(
            location=str(
                data_dir.joinpath("library/features/carriers_python/v1")
            ),
            import_module_name='.calculate',
            parameters={"foo": "bar"},
            progress_logger=progress_logger
        )

    # Assert
    assert isinstance(result, FeatureCarrierPythonTransformer)

    # make sure we can call transform on it
    df = create_empty_dataframe(spark_session=spark_session)
    result.transform(df)



from pathlib import Path

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_select_columns_transformer.v1.framework_select_columns_transformer import \
    FrameworkSelectColumnsTransformer
from tests.conftest import clean_spark_session

from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import FrameworkCsvLoader


# noinspection SqlNoDataSourceInspection
def test_can_keep_columns(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath('./')
    test_file_path: str = f"{data_dir.joinpath('test.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkCsvLoader(
        view="my_view", path_to_csv=test_file_path, delimiter=","
    ).transform(df)

    FrameworkSelectColumnsTransformer(
        view="my_view", keep_columns=['Column2']
    ).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM my_view")

    result.show()

    # Assert
    assert len(result.columns) == 1

    assert result.count() == 3

    assert result.collect()[1][0] == "bar"



from typing import Any, Dict, List

from spark_auto_mapper.automappers.automapper import AutoMapper
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase
from spark_auto_mapper.data_types.list import AutoMapperList
from spark_auto_mapper.helpers.automapper_helpers import AutoMapperHelpers as A


def mapping(parameters: Dict[str, Any]) -> List[AutoMapperBase]:
    # example of a variable
    client_address_variable: str = "address1"
    mapper = AutoMapper(
        view=parameters["view"], source_view="patients", keys=["member_id"]
    ).columns(
        patient_id=A.column("member_id"),
        dst1="src1",
        dst2=AutoMapperList([client_address_variable]),
        dst3=AutoMapperList([client_address_variable, "address2"]),
        dst4=AutoMapperList(
            [A.complex(use="usual", family=A.column("last_name"))]
        )
    )

    company_name: str = "Microsoft"

    if company_name == "Microsoft":
        mapper = mapper.columns(
            dst5=AutoMapperList(
                [A.complex(use="usual", family=A.column("last_name"))]
            )
        )

    mapper2 = AutoMapper(
        view=parameters["view2"], source_view="patients", keys=["member_id"]
    ).columns(
        patient_id=A.column("member_id"),
        dst1="src2",
        dst22=AutoMapperList([client_address_variable]),
    )

    return [mapper, mapper2]



from os import path, makedirs, listdir
from pathlib import Path
from shutil import rmtree, copyfile
from typing import Callable, List, Optional, Set

from spark_pipeline_framework.proxy_generator.proxy_generator import ProxyGenerator


def test_can_generate_proxies() -> None:
    data_dir: Path = Path(__file__).parent.joinpath('./')

    temp_folder = data_dir.joinpath('temp')
    if path.isdir(temp_folder):
        rmtree(temp_folder)

    makedirs(temp_folder)

    src_library_folder: str = path.join(data_dir, 'library')
    temp_library_folder: str = path.join(data_dir, 'temp', 'library')

    if not path.exists(temp_library_folder):
        makedirs(temp_library_folder)

    # copytree(src=src_library_folder, dst=temp_library_folder)
    recursive_overwrite(src=src_library_folder, dst=temp_library_folder)

    # Act
    ProxyGenerator.generate_proxies(folder=temp_library_folder)

    # Assert
    data_source_path = path.join(
        temp_library_folder, "data_sources", "my_data_source", "v1",
        "data_sources_my_data_source_v1.py"
    )
    assert (path.exists(data_source_path))

    assert (
        path.join(
            temp_library_folder, "features", "my_sql_feature", "v1",
            "features_my_sql_feature_v1.py"
        )
    )
    assert (
        path.join(
            temp_library_folder, "features", "my_python_feature", "v1",
            "features_my_python_feature_v1.py"
        )
    )


def recursive_overwrite(
    src: str,
    dst: str,
    ignore: Optional[Callable[[str, List[str]], Set[str]]] = None
) -> None:
    if path.isdir(src):
        print(f"dir={src}")
        if not path.isdir(dst):
            makedirs(dst)
        files = listdir(src)
        if ignore is not None:
            ignored = ignore(src, files)
        else:
            ignored = set()
        for f in files:
            print(f"file={f}")
            if f not in ignored:
                recursive_overwrite(
                    path.join(src, f), path.join(dst, f), ignore
                )
    else:
        print(f"src={src}, dst={dst}")
        copyfile(src, dst)



from pathlib import Path
from typing import Any, Dict

from pyspark.ml.pipeline import Pipeline
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from library.features.carriers.v1.features_carriers_v1 import FeaturesCarriersV1
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import FrameworkCsvLoader
from spark_pipeline_framework.utilities.pipeline_helper import create_steps


def test_simple_csv_and_sql_pipeline(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath('./')
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    parameters: Dict[str, Any] = {}

    stages = create_steps(
        [
            FrameworkCsvLoader(view="flights", path_to_csv=flights_path),
            FeaturesCarriersV1(parameters=parameters).transformers,
        ]
    )

    pipeline: Pipeline = Pipeline(stages=stages)
    transformer = pipeline.fit(df)
    transformer.transform(df)

    # Assert
    result_df: DataFrame = spark_session.sql("SELECT * FROM flights2")
    result_df.show()

    assert result_df.count() > 0



from pathlib import Path
from typing import Any, Dict, Union

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.transformers.framework_base_exporter.v1.framework_base_exporter import FrameworkBaseExporter


class FrameworkParquetExporter(FrameworkBaseExporter):

    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(self, file_path: Union[str, Path], **kwargs: Dict[Any, Any]):
        super().__init__(**kwargs)

        assert isinstance(file_path, Path) or isinstance(file_path, str)
        assert file_path

        self.logger = get_logger(__name__)

        self.file_path: Param = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self._set(file_path=file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setFilePath(
        self, value: Union[str, Path]
    ) -> 'FrameworkParquetExporter':
        self._paramMap[self.file_path] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[str, Path]:
        return self.getOrDefault(self.file_path)  # type: ignore

    def getFormat(self) -> str:
        return "parquet"

    def getOptions(self) -> Dict[str, Any]:
        return {"path": self.getFilePath()}



from typing import Any


def get_first_class_in_module(module: Any) -> Any:
    """
    Gets the first class in the specified module
    :param module:
    :return:
    """
    md = module.__dict__
    return [
        md[c] for c in md
        if (isinstance(md[c], type) and md[c].__module__ == module.__name__)
    ][0]



from typing import Any, Dict

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.transformers.framework_base_exporter.v1.framework_base_exporter import \
    FrameworkBaseExporter


class FrameworkJdbcExporter(FrameworkBaseExporter):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self, jdbc_url: str, table: str, driver: str, **kwargs: Dict[Any, Any]
    ):
        super().__init__(**kwargs)
        assert jdbc_url
        assert table
        assert driver

        self.logger = get_logger(__name__)

        self.jdbc_url: Param = Param(self, "jdbc_url", "")
        self._setDefault(jdbc_url=jdbc_url)

        self.table: Param = Param(self, "table", "")
        self._setDefault(table=table)

        self.driver: Param = Param(self, "driver", "")
        self._setDefault(driver=driver)

        self._set(jdbc_url=jdbc_url, table=table, driver=driver)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setJdbcUrl(self, value: Param) -> 'FrameworkJdbcExporter':
        self._paramMap[self.jdbc_url] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getJdbcUrl(self) -> str:
        return self.getOrDefault(self.jdbc_url)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setTable(self, value: Param) -> 'FrameworkJdbcExporter':
        self._paramMap[self.table] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getTable(self) -> str:
        return self.getOrDefault(self.table)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setDriver(self, value: Param) -> 'FrameworkJdbcExporter':
        self._paramMap[self.driver] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDriver(self) -> str:
        return self.getOrDefault(self.driver)  # type: ignore

    def getFormat(self) -> str:
        return "jdbc"

    def getOptions(self) -> Dict[str, Any]:
        return {
            "url": self.getJdbcUrl(),
            "dbtable": self.getTable(),
            "driver": self.getDriver(),
        }



from typing import List, Optional, Dict, Any, Union

from pyspark.ml.base import Transformer
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import ProgressLogMetric
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.FriendlySparkException import FriendlySparkException
from spark_pipeline_framework.utilities.pipeline_helper import create_steps


class FrameworkPipeline(Transformer):  # type: ignore
    def __init__(
        self, parameters: Dict[str, Any], progress_logger: ProgressLogger
    ):
        super(FrameworkPipeline, self).__init__()
        self.transformers: List[Transformer] = []
        self.parameters: Dict[str, Any] = parameters
        self.progress_logger: ProgressLogger = progress_logger

    # noinspection PyUnusedLocal
    def fit(self, df: DataFrame) -> 'FrameworkPipeline':
        return self

    def _transform(self, df: DataFrame) -> DataFrame:
        logger = get_logger(__name__)
        count_of_transformers: int = len(self.transformers)
        i: int = 0
        for transformer in self.transformers:
            stage_name: Optional[str] = None
            try:
                i += 1
                if hasattr(transformer, "getName"):
                    # noinspection Mypy
                    stage_name = transformer.getName()
                    logger.info(
                        f"---- Running transformer {stage_name}  ({i} of {count_of_transformers}) ----"
                    )
                    # self.spark_session.sparkContext.setJobDescription(stage_name)
                    # print_memory_stats(sc(df))
                with ProgressLogMetric(
                    progress_logger=self.progress_logger,
                    name=stage_name or "unknown"
                ):
                    df = transformer.transform(dataset=df)
            except Exception as e:
                logger.warning("======== stage threw exception =======")
                if hasattr(transformer, "getSql"):
                    # noinspection Mypy
                    logger.info(transformer.getSql())
                logger.warning("======== stage threw exception =======")
                # use exception chaining to add stage name but keep original exception
                raise FriendlySparkException(str(e), stage_name=stage_name)
        return df

    # noinspection PyMethodMayBeStatic
    def create_steps(
        self, my_list: List[Union[Transformer, List[Transformer]]]
    ) -> List[Transformer]:
        return create_steps(my_list)



from os import listdir, path
from pathlib import Path
from typing import Optional, List, Dict, Any, Union

from pyspark.ml.base import Transformer
from pyspark.sql import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.python_transformer_helpers import get_python_transformer_from_location, \
    get_python_function_from_location
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import FrameworkCsvLoader
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import FrameworkMappingLoader
from spark_pipeline_framework.transformers.framework_sql_transformer.v1.framework_sql_transformer import FrameworkSqlTransformer
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class ProxyBase(FrameworkTransformer):
    loader: Optional[Transformer] = None
    converter: Optional[Transformer] = None
    feature: Optional[Transformer] = None
    location: str = ""

    def __init__(
        self,
        parameters: Dict[str, Any],
        location: Union[str, Path],
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        super().__init__(
            name=self.__class__.__name__,
            parameters=parameters,
            progress_logger=progress_logger
        )
        self.verify_count_remains_same: bool = verify_count_remains_same
        self.location: str = str(location)

        assert self.location
        # Iterate over files to create transformers
        files: List[str] = listdir(self.location)
        index_of_module: int = self.location.rfind("/library/")
        module_name: str = self.location[index_of_module +
                                         1:].replace("/", ".")

        # noinspection Mypy
        self._set(
            name=self.__class__.__name__,
            parameters=parameters,
            progress_logger=progress_logger
        )

        for file in files:
            if file == 'my_view.sql':
                load_sql: str = self.read_file_as_string(
                    path.join(self.location, file)
                ).format(parameters=parameters)
                self.loader = FrameworkSqlTransformer(
                    sql=load_sql,
                    name=module_name,
                    progress_logger=progress_logger,
                    log_sql=parameters.get("debug_log_sql", False)
                )
            elif file.endswith('.csv') and self.loader is None:
                file_name = file.replace('.csv', '')
                self.loader = FrameworkCsvLoader(
                    view=file_name,
                    path_to_csv=path.join(self.location, file),
                    delimiter=parameters.get("delimiter", ",")
                )
            elif file == 'convert.sql':
                convert_sql: str = self.read_file_as_string(path.join(self.location, file)) \
                    .format(parameters=parameters)
                self.converter = FrameworkSqlTransformer(
                    sql=convert_sql,
                    name=module_name,
                    progress_logger=progress_logger,
                    log_sql=parameters.get("debug_log_sql", False)
                )
            elif file.endswith(
                '.sql'
            ) and self.loader is None and self.converter is None:
                feature_sql: str = self.read_file_as_string(path.join(self.location, file)) \
                    .format(parameters=parameters)
                self.feature = FrameworkSqlTransformer(
                    sql=feature_sql,
                    name=module_name,
                    progress_logger=progress_logger,
                    log_sql=parameters.get("debug_log_sql", False),
                    view=file.replace('.sql', ''),
                    verify_count_remains_same=verify_count_remains_same
                )
            elif file == 'calculate.py':
                self.feature = self.get_python_transformer('.calculate')
            elif file == 'mapping.py':
                self.feature = self.get_python_mapping_transformer('.mapping')

    @staticmethod
    def read_file_as_string(file_path: str) -> str:
        with open(file_path, 'r') as file:
            file_contents = file.read()
        return file_contents

    @property
    def transformers(self) -> List[Transformer]:
        return [
            transformer
            for transformer in [self.loader, self.converter, self.feature]
            if transformer is not None
        ]

    @transformers.setter
    def transformers(self, value: Any) -> None:
        raise AttributeError("transformers property is read only.")

    def _transform(self, df: DataFrame) -> DataFrame:
        # iterate through my transformers
        transformer: Transformer
        for transformer in self.transformers:
            df = transformer.transform(df)
        return df

    def get_python_transformer(self, import_module_name: str) -> Transformer:
        progress_logger = self.getProgressLogger()
        assert progress_logger
        return get_python_transformer_from_location(
            location=self.location,
            import_module_name=import_module_name,
            parameters=self.getParameters() or {},
            progress_logger=progress_logger
        )

    def get_python_mapping_transformer(
        self, import_module_name: str
    ) -> Transformer:
        parameters: Optional[Dict[str, Any]] = self.getParameters()
        return FrameworkMappingLoader(
            view=parameters["view"]
            if parameters and "view" in parameters else "",
            mapping_function=get_python_function_from_location(
                location=self.location, import_module_name=import_module_name
            ),
            parameters=parameters,
            progress_logger=self.getProgressLogger()
        )



from pathlib import Path
from typing import Union, List, Dict, Any

from pyspark import keyword_only
from pyspark.ml.param.shared import Param

from spark_pipeline_framework.transformers.framework_local_file_loader.v1.framework_local_file_loader import \
    FrameworkLocalFileLoader


class FrameworkJsonLoader(FrameworkLocalFileLoader):

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    @keyword_only
    def __init__(
        self,
        view: str,
        filepath: Union[str, List[str], Path],
        multiLine: bool = False,
        **kwargs: Dict[Any, Any]
    ):
        super().__init__(view=view, filepath=filepath, **kwargs)

        self.multiLine: Param = Param(self, "multiLine", "")
        self._setDefault(multiLine=False)
        self._set(multiLine=multiLine)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setMultiLine(self, value: Param) -> 'FrameworkJsonLoader':
        self._paramMap[self.multiLine] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiLine(self) -> str:
        return self.getOrDefault(self.multiLine)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderFormat(self) -> str:
        return "json"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = {"multiLine": self.getMultiLine()}
        return options



import json
from glob import glob
from os import path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import List

from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_json_splitter.v1.framework_json_splitter import \
    FrameworkJsonSplitter
from spark_pipeline_framework.utilities.spark_data_frame_helpers import create_empty_dataframe


def test_json_splitter(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath('./')

    temp_folder = data_dir.joinpath('./temp')
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    insurance_feed_path: Path = data_dir.joinpath("large_json_file.json")

    # Act
    with ProgressLogger() as progress_logger:
        FrameworkJsonSplitter(
            file_path=insurance_feed_path,
            output_folder=temp_folder,
            max_size_per_file_in_mb=0.1,
            progress_logger=progress_logger
        ).transform(df)

    # Assert
    files: List[str] = glob(str(temp_folder.joinpath("*.json")))
    assert len(files) == 14

    # read the last file and make sure it is valid json

    last_file: Path = temp_folder.joinpath("large_json_file_6.json")
    with open(last_file, "r+") as file:
        obj = json.loads(file.read())
        assert obj[0]["name"] == "Sellers Mcguire"
        print(obj)



from typing import List, Union

from pyspark.ml import Transformer

from spark_pipeline_framework.utilities.flattener import flatten


def create_steps(
    my_list: List[Union[Transformer, List[Transformer]]]
) -> List[Transformer]:
    return flatten(my_list)



from os import path
from pathlib import Path

from pyspark.sql.session import SparkSession


class ParquetHelper:
    @staticmethod
    def create_parquet_from_csv(
        spark_session: SparkSession,
        file_path: str,
        delimiter: str = ",",
        has_header: bool = True,
        infer_schema: bool = True
    ) -> str:
        data_dir: Path = Path(file_path).parent
        csv_path: str = f"file://{file_path}"
        parquet_file_name: str = path.splitext(path.basename(file_path))[0]
        parquet_full_path: str = f"file://{data_dir.joinpath('temp/').joinpath(f'{parquet_file_name}.parquet')}"

        ParquetHelper.save_csv_as_parquet(
            spark_session, csv_path, parquet_full_path, delimiter, has_header,
            infer_schema
        )

        return parquet_full_path

    @staticmethod
    def save_csv_as_parquet(
        spark_session: SparkSession,
        source_path: str,
        destination_path: str,
        delimiter: str = ",",
        has_header: bool = True,
        infer_schema: bool = False
    ) -> None:
        # noinspection SpellCheckingInspection
        spark_session.read \
            .option("delimiter", delimiter) \
            .option("header", str(has_header)) \
            .option("inferschema", infer_schema) \
            .csv(source_path) \
            .coalesce(1) \
            .write \
            .format("parquet") \
            .mode("overwrite") \
            .save(destination_path)



from typing import Optional, List, Dict, Any

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class FrameworkSelectColumnsTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: Optional[str] = None,
        drop_columns: Optional[List[str]] = None,
        keep_columns: Optional[List[str]] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.logger = get_logger(__name__)

        if not view:
            raise ValueError("view is None or empty")

        self.view: Param = Param(self, "view", "")
        self._setDefault(view=None)

        self.drop_columns: Param = Param(self, "drop_columns", "")
        self._setDefault(drop_columns=None)

        self.keep_columns: Param = Param(self, "keep_columns", "")
        self._setDefault(keep_columns=None)

        self.verify_count_remains_same: Param = Param(
            self, "verify_count_remains_same", ""
        )
        self._setDefault(verify_count_remains_same=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyUnusedLocal,PyMissingOrEmptyDocstring,PyPep8Naming
    @keyword_only
    def setParams(
        self,
        view: Optional[str] = None,
        drop_columns: Optional[List[str]] = None,
        keep_columns: Optional[List[str]] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False
    ) -> None:
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        return self._set(**kwargs)  # type: ignore

    def _transform(self, df: DataFrame) -> DataFrame:
        # name = self.getName()
        view = self.getView()
        # progress_logger: ProgressLogger = self.getProgressLogger()
        drop_columns: Optional[List[str]] = self.getOrDefault("drop_columns")
        keep_columns: Optional[List[str]] = self.getOrDefault("keep_columns")

        result_df: DataFrame = df.sql_ctx.table(view)

        if keep_columns and len(keep_columns) > 0:
            all_columns = result_df.columns
            drop_columns = list(set(all_columns).difference(set(keep_columns)))

        if drop_columns and len(drop_columns) > 0:
            self.logger.info(
                f"FrameworkSelectColumnsTransformer: Dropping columns {drop_columns} from {view}"
            )
            result_df.drop(*drop_columns).createOrReplaceTempView(view)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setSql(self, value: str) -> 'FrameworkSelectColumnsTransformer':
        self._paramMap[self.sql] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSql(self) -> str:
        return self.getOrDefault(self.sql)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setView(self, value: str) -> 'FrameworkSelectColumnsTransformer':
        self._paramMap[self.view] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setLogSql(self, value: str) -> 'FrameworkSelectColumnsTransformer':
        self._paramMap[self.log_sql] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLogSql(self) -> str:
        return self.getOrDefault(self.log_sql)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setVerifyCountRemainsSame(
        self, value: bool
    ) -> 'FrameworkSelectColumnsTransformer':
        self._paramMap[self.verify_count_remains_same] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getVerifyCountRemainsSame(self) -> bool:
        return self.getOrDefault(  # type: ignore
            self.verify_count_remains_same
        )



from os import path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import List

from spark_pipeline_framework.utilities.json_to_jsonl_converter import convert_json_to_jsonl


def test_can_convert_json_object_to_jsonl() -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath('./')

    temp_folder = data_dir.joinpath('./temp')
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    # Act
    output_file = convert_json_to_jsonl(
        src_file=data_dir.joinpath("json_file.json"),
        dst_file=temp_folder.joinpath("json_file.json")
    )

    # Assert
    with open(output_file, "r+") as file:
        lines: List[str] = file.readlines()
        assert len(lines) == 1
        assert lines[0] == '{"foo":"bar","foo2":"bar2"}\n'



import json
import math
import os
import shutil
from pathlib import Path
from typing import Dict, Any, Optional, Union, List

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import FrameworkTransformer


class FrameworkJsonSplitter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        file_path: Union[Path, str],
        output_folder: Union[Path, str],
        max_size_per_file_in_mb: float,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ):
        """
        Splits a large json file into multiple files where the maximum size < max_size_per_file_in_mb
        :param file_path: path to original json file
        :param output_folder: folder to put the splitted json files
        :param max_size_per_file_in_mb:
        :param name:
        :param parameters:
        :param progress_logger:
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.file_path: Param = Param(self, "file_path", "")
        self._setDefault(file_path=file_path)

        self.output_folder: Param = Param(self, "output_folder", "")
        self._setDefault(output_folder=output_folder)

        self.max_size_per_file_in_mb: Param = Param(
            self, "max_size_per_file_in_mb", ""
        )
        self._setDefault(max_size_per_file_in_mb=max_size_per_file_in_mb)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    @keyword_only
    def setParams(
        self,
        file_path: Union[Path, str],
        output_folder: Union[Path, str],
        max_size_per_file_in_mb: float,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None
    ) -> Any:
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        file_path: Union[Path, str] = self.getFilePath()
        max_size_per_file_in_mb: float = self.getMaxSizePerFileInMb()
        output_folder: Union[Path, str] = self.getOutputFolder()
        try:
            # request file file_name
            with open(file_path, "r") as f:
                file_size = os.path.getsize(file_path)
                data = json.load(f)

            if isinstance(data, list):
                data_len = len(data)
                # check that file is larger than max size
                if file_size < max_size_per_file_in_mb * 1000000:
                    print('File smaller than split size, exiting')
                    # just write the whole file
                    shutil.copyfile(str(file_path), str(output_folder))
                else:
                    # determine number of files necessary
                    num_files = math.ceil(
                        file_size / (max_size_per_file_in_mb * 1000000)
                    )
                    self.logger.info(
                        f'File will be split into {num_files} equal parts'
                    )
                    # initialize 2D array
                    split_data: List[List[Any]
                                     ] = [[] for i in range(0, num_files)]
                    # determine indices of cutoffs in array
                    starts = [
                        math.floor(i * data_len / num_files)
                        for i in range(0, num_files)
                    ]
                    starts.append(data_len)
                    # loop through 2D array
                    for i in range(0, num_files):
                        # loop through each range in array
                        for n in range(starts[i], starts[i + 1]):
                            split_data[i].append(data[n])

                        # create file when section is complete
                        file_name: str = os.path.basename(file_path).split(
                            '.'
                        )[0] + '_' + str(i + 1) + '.json'
                        with open(
                            Path(output_folder).joinpath(file_name), 'w+'
                        ) as outfile:
                            json.dump(split_data[i], outfile)
                        self.logger.info(
                            f'Part{i + 1}... completed: {Path(output_folder).joinpath(file_name)}'
                        )
            else:
                raise Exception("Not valid JSON")
        except Exception as e:
            self.logger.error(f"File read failed for {file_path}")
            raise e

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[Path, str]:
        return self.getOrDefault(self.file_path)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOutputFolder(self) -> Union[Path, str]:
        return self.getOrDefault(self.output_folder)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMaxSizePerFileInMb(self) -> float:
        return self.getOrDefault(self.max_size_per_file_in_mb)  # type: ignore



from pathlib import Path
from typing import Dict, Any

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from library.features.carriers.v1.features_carriers_v1 import FeaturesCarriersV1
from library.features.carriers_python.v1.features_carriers_python_v1 import FeaturesCarriersPythonV1
from spark_pipeline_framework.pipelines.framework_pipeline import FrameworkPipeline
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import FrameworkCsvLoader


class MyPipeline(FrameworkPipeline):
    def __init__(
        self, parameters: Dict[str, Any], progress_logger: ProgressLogger
    ):
        super(MyPipeline, self).__init__(
            parameters=parameters, progress_logger=progress_logger
        )
        self.transformers = self.create_steps(
            [
                FrameworkCsvLoader(
                    view="flights",
                    path_to_csv=parameters["flights_path"],
                    parameters=parameters,
                    progress_logger=progress_logger
                ),
                FeaturesCarriersV1(
                    parameters=parameters, progress_logger=progress_logger
                ).transformers,
                FeaturesCarriersPythonV1(
                    parameters=parameters, progress_logger=progress_logger
                ).transformers
            ]
        )


def test_can_run_framework_pipeline(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath('./')
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    parameters = {"flights_path": flights_path}

    with ProgressLogger() as progress_logger:
        pipeline: MyPipeline = MyPipeline(
            parameters=parameters, progress_logger=progress_logger
        )
        transformer = pipeline.fit(df)
        transformer.transform(df)

    # Assert
    result_df: DataFrame = spark_session.sql("SELECT * FROM flights2")
    result_df.show()

    assert result_df.count() > 0



# noinspection Mypy
from typing import List, Any

from setuptools import setup, find_packages
from os import path, getcwd

# from https://packaging.python.org/tutorials/packaging-projects/

# noinspection SpellCheckingInspection
package_name = 'sparkpipelineframework'

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    with open(path.join(getcwd(), 'VERSION')) as version_file:
        version = version_file.read().strip()
except IOError:
    raise


def fix_setuptools() -> None:
    """Work around bugs in setuptools.

    Some versions of setuptools are broken and raise SandboxViolation for normal
    operations in a virtualenv. We therefore disable the sandbox to avoid these
    issues.
    """
    try:
        from setuptools.sandbox import DirectorySandbox

        # noinspection PyUnusedLocal
        def violation(operation: Any, *args: Any, **_: Any) -> None:
            print("SandboxViolation: %s" % (args, ))

        DirectorySandbox._violation = violation
    except ImportError:
        pass


# Fix bugs in setuptools.
fix_setuptools()


def parse_requirements(file: str) -> List[str]:
    with open(file, "r") as fs:
        return [
            r for r in fs.read().splitlines() if (
                len(r.strip()) > 0 and not r.strip().startswith("#")
                and not r.strip().startswith("--")
            )
        ]


requirements: List[str] = parse_requirements('requirements.txt')
test_requirements: List[str] = parse_requirements('requirements-test.txt')

# classifiers list is here: https://pypi.org/classifiers/

# create the package setup
setup(
    name=package_name,
    version=version,
    author="Imran Qureshi",
    author_email="imranq2@hotmail.com",
    description="Framework for simpler Spark Pipelines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/imranq2/SparkPipelineFramework",
    packages=find_packages(),
    install_requires=requirements,
    tests_require=test_requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    dependency_links=[],
    include_package_data=True,
    zip_safe=False,
    package_data={"spark_pipeline_framework": ["py.typed"]},
    data_files=[
        'spark_pipeline_framework/Makefile.spark',
        'spark_pipeline_framework/Makefile.docker'
    ]
)



# noinspection DuplicatedCode
from sys import stderr
from logging import Logger, getLogger, StreamHandler, Formatter, INFO


def get_logger(name: str, level: int = INFO) -> Logger:
    logger: Logger = getLogger(name)
    logger.setLevel(level)

    if logger.handlers:
        pass
    else:
        stream_handler: StreamHandler = StreamHandler(stderr)
        stream_handler.setLevel(level=level)
        # noinspection SpellCheckingInspection
        formatter: Formatter = Formatter(
            '%(asctime)s.%(msecs)03d %(levelname)s %(module)s %(lineno)d - %(funcName)s: %(message)s'
        )
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    return logger



from pathlib import Path
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Optional

from spark_pipeline_framework.logger.yarn_logger import get_logger


class ProgressLogger:
    def __init__(self) -> None:
        self.logger = get_logger(__name__)

    def __enter__(self) -> 'ProgressLogger':
        return self

    def __exit__(
        self, exc_type: Optional[BaseException],
        exc_value: Optional[BaseException], traceback: Optional[TracebackType]
    ) -> None:
        pass

    def log_metric(self, name: str, time_diff_in_minutes: float) -> None:
        self.logger.info(f"{name}: {time_diff_in_minutes} min")

    # noinspection PyUnusedLocal
    def log_artifact(
        self,
        key: str,
        contents: str,
        folder_path: Optional[str] = None
    ) -> None:
        try:
            with TemporaryDirectory() as temp_dir_name:
                data_dir: Path = Path(temp_dir_name)
                file_path: Path = data_dir.joinpath(key)
                with open(file_path, 'w') as file:
                    file.write(contents)
                    self.logger.info(f"Wrote sql to {file_path}")

        except Exception as e:
            self.logger.warning(
                f"Error in log_artifact writing to mlflow: {str(e)}"
            )

    def write_to_log(self, name: str, message: str = "") -> bool:
        self.logger.info(name + ": " + str(message))
        return True

    def log_exception(self, e: Exception) -> None:
        self.log_artifact("_exception.txt", str(e))
